// The class IDs of these Shell extension class.
//
// class ids:
//
// D14BEDD3-4E05-4F2F-B0DE-D0381E6AE606 (seafile-client uses C0381E6AE606)
//
//

DEFINE_GUID(CLSID_SEAFILE_SHELLEXT,             0xd14bedd3, 0x4e05, 0x4f2f, 0xb0, 0xde, 0xd0, 0x38, 0x1e, 0x6a, 0xe6, 0x06);

DEFINE_GUID(CLSID_SEAFILE_ICON_SYNCING,         0xd14bedd3, 0x4e05, 0x4f2f, 0xb0, 0xde, 0xd0, 0x38, 0x1e, 0x6a, 0xe6, 0x07);

DEFINE_GUID(CLSID_SEAFILE_ICON_ERROR,           0xd14bedd3, 0x4e05, 0x4f2f, 0xb0, 0xde, 0xd0, 0x38, 0x1e, 0x6a, 0xe6, 0x08);

DEFINE_GUID(CLSID_SEAFILE_ICON_SYNCED,          0xd14bedd3, 0x4e05, 0x4f2f, 0xb0, 0xde, 0xd0, 0x38, 0x1e, 0x6a, 0xe6, 0x09);

DEFINE_GUID(CLSID_SEAFILE_ICON_PARTIAL_SYNCED,  0xd14bedd3, 0x4e05, 0x4f2f, 0xb0, 0xde, 0xd0, 0x38, 0x1e, 0x6a, 0xe6, 0x0a);

DEFINE_GUID(CLSID_SEAFILE_ICON_CLOUD,           0xd14bedd3, 0x4e05, 0x4f2f, 0xb0, 0xde, 0xd0, 0x38, 0x1e, 0x6a, 0xe6, 0x0b);

DEFINE_GUID(CLSID_SEAFILE_ICON_READ_ONLY,       0xd14bedd3, 0x4e05, 0x4f2f, 0xb0, 0xde, 0xd0, 0x38, 0x1e, 0x6a, 0xe6, 0x0c);

DEFINE_GUID(CLSID_SEAFILE_ICON_LOCKED_BY_OTHERS,0xd14bedd3, 0x4e05, 0x4f2f, 0xb0, 0xde, 0xd0, 0x38, 0x1e, 0x6a, 0xe6, 0x0d);

DEFINE_GUID(CLSID_SEAFILE_ICON_LOCKED_BY_ME,    0xd14bedd3, 0x4e05, 0x4f2f, 0xb0, 0xde, 0xd0, 0x38, 0x1e, 0x6a, 0xe6, 0x0e);
