QToolButton#mDownloadTasksBtn,
QToolButton#mServerStatusBtn
{
    border-style:flat;
}

QToolButton:hover
{
    border: 1px solid #ccc;
    border-radius: 4px;
    background-color: rgb(255,255,255);
}

QWidget#mainWrapper
{
    border : 0;
    border-radius: 0px;
}

QWidget#mAccountWrapper {
    margin-top: 0px;
}

/*
DownloadRepoDialog {
    min-height: 400px;
    max-height: 400px;
}
*/

QToolButton#mAccountBtn
{
    margin-right: 0px;
}


QFrame#mFrame
{
    border: 0;
}

FileBrowserDialog QWidget#mainWidget
{
    border : 0;
    border-radius: 0px;
}

FileBrowserDialog QToolBar QToolButton#backwardButton {
    margin-right: 0px;
}

FileBrowserDialog QToolBar QToolButton#forwardButton {
    margin-left: 0px;
}

QMessageBox {
    messagebox-text-interaction-flags: NoTextInteraction;
}
